/* code6804.c */
/*****************************************************************************/
/* AS-Portierung                                                             */
/*                                                                           */
/* AS-Codeenerator Motorola/ST 6804                                          */
/*                                                                           */
/* Historie: 17.10.1996 Grundsteinlegung                                     */
/*            2. 1.1999 ChkPC-Anpassung                                      */
/*            9. 3.2000 'ambiguous else'-Warnungen beseitigt                 */
/*                                                                           */
/*****************************************************************************/
/* $Id: code6804.c,v 1.5 2014/11/13 14:08:32 alfred Exp $                    */
/*****************************************************************************
 * $Log: code6804.c,v $
 * Revision 1.5  2014/11/13 14:08:32  alfred
 * - rework to current style
 *
 * Revision 1.4  2014/06/09 12:46:51  alfred
 * - add missing parentheses
 *
 * Revision 1.3  2005/09/08 17:31:03  alfred
 * - add missing include
 *
 * Revision 1.2  2004/05/29 12:04:46  alfred
 * - relocated DecodeMot(16)Pseudo into separate module
 *
 *****************************************************************************/

#include "stdinc.h"

#include <string.h>

#include "bpemu.h"
#include "strutil.h"
#include "asmdef.h"
#include "asmsub.h"
#include "asmpars.h"
#include "asmitree.h"
#include "codepseudo.h"
#include "motpseudo.h"
#include "codevars.h"

#include "code6804.h"

typedef struct
{
  char *Name;
  LongInt Code;
} BaseOrder;

#define FixedOrderCnt 19

enum
{
  ModNone = -1,
  ModInd = 0,
  ModDir = 1,
  ModImm = 2,
};

#define MModInd (1 << ModInd)
#define MModDir (1 << ModDir)
#define MModImm (1 << ModImm)

static ShortInt AdrMode;
static Byte AdrVal;

static CPUVar CPU6804;

static BaseOrder *FixedOrders;

/*--------------------------------------------------------------------------*/

static void DecodeAdr(char *Asc, Boolean MayImm)
{
  Boolean OK;

  AdrMode = ModNone;

  if (!strcasecmp(Asc, "(X)"))
  {
    AdrMode = ModInd;
    AdrVal = 0x00;
    goto chk;
  }
  if (!strcasecmp(Asc, "(Y)"))
  {
    AdrMode = ModInd;
    AdrVal = 0x10;
    goto chk;
  }

  if (*Asc == '#')
  {
    AdrVal = EvalIntExpression(Asc + 1, Int8, &OK);
    if (OK)
      AdrMode = ModImm;
    goto chk;
  }

  AdrVal = EvalIntExpression(Asc, Int8, &OK);
  if (OK)
    AdrMode = ModDir;

chk:
  if ((AdrMode == ModImm) && (!MayImm))
  {
    WrError(1350); AdrMode = ModNone;
  }
}

static Boolean IsShort(Byte Adr)
{
  return ((Adr & 0xfc) == 0x80);
}

/*--------------------------------------------------------------------------*/

/* Anweisungen ohne Argument */

static void DecodeFixed(Word Index)
{
  const BaseOrder *pOrder = FixedOrders + Index;

  if (ArgCnt != 0) WrError(1110);
  else
  {
    if ((pOrder->Code >> 16) != 0)
      CodeLen = 3;
    else
      CodeLen = 1 + Ord(Hi(pOrder->Code) != 0);
    if (CodeLen == 3)
      BAsmCode[0] = pOrder->Code >> 16;
    if (CodeLen >= 2)
      BAsmCode[CodeLen - 2] = Hi(pOrder->Code);
    BAsmCode[CodeLen - 1] = Lo(pOrder->Code);
  }
}

/* relative/absolute Spruenge */

static void DecodeRel(Word Code)
{
  if (ArgCnt != 1) WrError(1110);
  else
  {
    Boolean OK;
    Integer AdrInt = EvalIntExpression(ArgStr[1], Int16, &OK) - (EProgCounter() + 1);

    if (OK)
    {
      if ((!SymbolQuestionable) && ((AdrInt < -16) || (AdrInt > 15))) WrError(1370);
      else
      {
        CodeLen = 1;
        BAsmCode[0] = Code + (AdrInt & 0x1f);
        ChkSpace(SegCode);
      }
    }
  }
}

static void DecodeJSR_JMP(Word Code)
{
  if (ArgCnt != 1) WrError(1110);
  else
  {
    Boolean OK;
    Word AdrInt = EvalIntExpression(ArgStr[1], UInt12, &OK);

    if (OK)
    {
      CodeLen = 2;
      BAsmCode[1] = Lo(AdrInt);
      BAsmCode[0] = Code + (Hi(AdrInt) & 15);
      ChkSpace(SegCode);
    }
  }
}

/* AKKU-Operationen */

static void DecodeALU(Word Code)
{
  if (ArgCnt != 1) WrError(1110);
  else
  {
    DecodeAdr(ArgStr[1], True);
    switch (AdrMode)
    {
      case ModInd:
        CodeLen = 1;
        BAsmCode[0] = 0xe0 + AdrVal + Code;
        break;
      case ModDir:
        CodeLen = 2;
        BAsmCode[0] = 0xf8 + Code;
        BAsmCode[1] = AdrVal;
        break;
      case ModImm:
        CodeLen = 2;
        BAsmCode[0] = 0xe8 + Code;
        BAsmCode[1] = AdrVal;
        break;
    }
  }
}

/* Datentransfer */

static void DecodeLDA_STA(Word Code)
{
  if (ArgCnt != 1) WrError(1110);
  else
  {
    DecodeAdr(ArgStr[1], !Code);
    switch (AdrMode)
    {
      case ModInd:
        CodeLen = 1;
        BAsmCode[0] = 0xe0 + Code + AdrVal;
        break;
      case ModDir:
        if (IsShort(AdrVal))
        {
          CodeLen = 1;
          BAsmCode[0] = 0xac + (Code << 4) + (AdrVal & 3);
        }
        else
        {
          CodeLen = 2;
          BAsmCode[0] = 0xf8 + Code;
          BAsmCode[1] = AdrVal;
        }
        break;
      case ModImm:
        CodeLen = 2;
        BAsmCode[0] = 0xe8 + Code;
        BAsmCode[1] = AdrVal;
        break;
    }
  }
}

static void DecodeLDXI_LDYI(Word Code)
{
  if (ArgCnt != 1) WrError(1110);
  else if (*ArgStr[1] != '#') WrError(1350);
  else
  {
    Boolean OK;

    BAsmCode[2] = EvalIntExpression(ArgStr[1] + 1, Int8, &OK);
    if (OK)
    {
      CodeLen = 3;
      BAsmCode[0] = 0xb0;
      BAsmCode[1] = Code;
    }
  }
}

static void DecodeMVI(Word Code)
{
  UNUSED(Code);

  if (ArgCnt != 2) WrError(1110);
  else if (*ArgStr[2] != '#') WrError(1350);
  else
  {
    Boolean OK;

    BAsmCode[1] = EvalIntExpression(ArgStr[1], Int8, &OK);
    if (OK)
    {
      ChkSpace(SegData);
      BAsmCode[2] = EvalIntExpression(ArgStr[2] + 1, Int8, &OK);
      if (OK)
      {
        BAsmCode[0] = 0xb0;
        CodeLen = 3;
      }
    }
  }
}

/* Read/Modify/Write-Operationen */

static void DecodeINC_DEC(Word Code)
{
  if (ArgCnt != 1) WrError(1110);
  else
  {
    DecodeAdr(ArgStr[1], False);
    switch (AdrMode)
    {
      case ModInd:
        CodeLen = 1;
        BAsmCode[0] = 0xe6 + Code + AdrVal;
        break;
      case ModDir:
        if (IsShort(AdrVal))
        {
          CodeLen = 1;
          BAsmCode[0] = 0xa8 + (Code << 4) + (AdrVal & 3);
        }
        else
        {
          CodeLen = 2;
          BAsmCode[0] = 0xfe + Code;
          BAsmCode[1] = AdrVal;
        }
        break;
    }
  }
}

/* Bitbefehle */

static void DecodeBSET_BCLR(Word Code)
{
  if (ArgCnt != 2) WrError(1110);
  else
  {
    Boolean OK;
    Byte Bit = EvalIntExpression(ArgStr[1], UInt3, &OK);
    if (OK)
    {
      BAsmCode[0] = Code + Bit;
      BAsmCode[1] = EvalIntExpression(ArgStr[2], Int8, &OK);
      if (OK)
      {
        CodeLen = 2;
        ChkSpace(SegData);
      }
    }
  }
}

static void DecodeBRSET_BRCLR(Word Code)
{
  if (ArgCnt != 3) WrError(1110);
  else
  {
    Boolean OK;
    Byte Bit = EvalIntExpression(ArgStr[1], UInt3, &OK);

    if (OK)
    {
      BAsmCode[0] = Code + Bit;
      BAsmCode[1] = EvalIntExpression(ArgStr[2], Int8, &OK);
      if (OK)
      {
        Integer AdrInt;

        ChkSpace(SegData);
        AdrInt = EvalIntExpression(ArgStr[3], Int16, &OK) - (EProgCounter() + 3);
        if (OK)
        {
          if ((!SymbolQuestionable) && ((AdrInt < -128) || (AdrInt > 127))) WrError(1370);
          else
          {
            ChkSpace(SegCode);
            BAsmCode[2] = AdrInt & 0xff;
            CodeLen = 3;
          }
        }
      }
    }
  }
}

static void DecodeSFR(Word Code)
{
  UNUSED(Code);

  CodeEquate(SegData, 0, 0xff);
}

/*--------------------------------------------------------------------------*/

static void AddFixed(char *NName, LongInt NCode)
{
  if (InstrZ >= FixedOrderCnt) exit(255);
  FixedOrders[InstrZ].Code = NCode;
  AddInstTable(InstTable, NName, InstrZ++, DecodeFixed);
}

static void AddRel(char *NName, LongInt NCode)   
{
  AddInstTable(InstTable, NName, NCode, DecodeRel);
}

static void AddALU(char *NName, LongInt NCode)   
{
  AddInstTable(InstTable, NName, NCode, DecodeALU);
}

static void InitFields(void)
{
  InstTable = CreateInstTable(203);
  AddInstTable(InstTable, "JMP", 0x90, DecodeJSR_JMP);
  AddInstTable(InstTable, "JSR", 0x80, DecodeJSR_JMP);
  AddInstTable(InstTable, "LDA", 0, DecodeLDA_STA);
  AddInstTable(InstTable, "STA", 1, DecodeLDA_STA);
  AddInstTable(InstTable, "LDXI", 0x80, DecodeLDXI_LDYI);
  AddInstTable(InstTable, "LDYI", 0x81, DecodeLDXI_LDYI);
  AddInstTable(InstTable, "MVI", 0, DecodeMVI);
  AddInstTable(InstTable, "INC", 0, DecodeINC_DEC);
  AddInstTable(InstTable, "DEC", 1, DecodeINC_DEC);
  AddInstTable(InstTable, "BSET", 0xd8, DecodeBSET_BCLR);
  AddInstTable(InstTable, "BCLR", 0xd0, DecodeBSET_BCLR);
  AddInstTable(InstTable, "BRSET", 0xc8, DecodeBRSET_BRCLR);
  AddInstTable(InstTable, "BRCLR", 0xc0, DecodeBRSET_BRCLR);
  AddInstTable(InstTable, "SFR", 0, DecodeSFR);

  FixedOrders = (BaseOrder *) malloc(sizeof(BaseOrder) * FixedOrderCnt); InstrZ = 0;
  AddFixed("CLRA", 0x00fbff);
  AddFixed("CLRX", 0xb08000);
  AddFixed("CLRY", 0xb08100);
  AddFixed("COMA", 0x0000b4);
  AddFixed("ROLA", 0x0000b5);
  AddFixed("ASLA", 0x00faff);
  AddFixed("INCA", 0x00feff);
  AddFixed("INCX", 0x0000a8);
  AddFixed("INCY", 0x0000a9);
  AddFixed("DECA", 0x00ffff);
  AddFixed("DECX", 0x0000b8);
  AddFixed("DECY", 0x0000b9);
  AddFixed("TAX" , 0x0000bc);
  AddFixed("TAY" , 0x0000bd);
  AddFixed("TXA" , 0x0000ac);
  AddFixed("TYA" , 0x0000ad);
  AddFixed("RTS" , 0x0000b3);
  AddFixed("RTI" , 0x0000b2);
  AddFixed("NOP" , 0x000020);

  AddRel("BCC", 0x40);
  AddRel("BHS", 0x40);
  AddRel("BCS", 0x60);
  AddRel("BLO", 0x60);
  AddRel("BNE", 0x00);
  AddRel("BEQ", 0x20);

  AddALU("ADD", 0x02);
  AddALU("SUB", 0x03);
  AddALU("CMP", 0x04);
  AddALU("AND", 0x05);
}

static void DeinitFields(void)
{
  DestroyInstTable(InstTable);

  free(FixedOrders);
}

/*--------------------------------------------------------------------------*/

static void MakeCode_6804(void)
{
  CodeLen = 0;
  DontPrint = False;

  /* zu ignorierendes */

  if (Memo(""))
    return;

  /* Pseudoanweisungen */

  if (DecodeMotoPseudo(True))
    return;

  if (!LookupInstTable(InstTable, OpPart))
    WrXError(1200, OpPart);
}

static Boolean IsDef_6804(void)
{
  return (Memo("SFR"));
}

static void SwitchFrom_6804(void)
{
  DeinitFields();
}

static void SwitchTo_6804(void)
{
  TurnWords = False;
  ConstMode = ConstModeMoto;
  SetIsOccupied = False;

  PCSymbol = "PC";
  HeaderID = 0x64;
  NOPCode = 0x20;
  DivideChars = ",";
  HasAttrs = False;

  ValidSegs = (1 << SegCode) | (1 << SegData);
  Grans[SegCode] = 1; ListGrans[SegCode] = 1; SegInits[SegCode] = 0;
  SegLimits[SegCode] = 0xfff;
  Grans[SegData] = 1; ListGrans[SegData] = 1; SegInits[SegData] = 0;
  SegLimits[SegData] = 0xff;

  MakeCode = MakeCode_6804;
  IsDef = IsDef_6804;
  SwitchFrom = SwitchFrom_6804;
  InitFields();
}

void code6804_init(void)
{
  CPU6804 = AddCPU("6804", SwitchTo_6804);
}
